'use strict';

var codegen = require('shift-codegen').default,
    traverse = require('shift-traverse');

var tree,
    stx = traverse.Syntax;

var ast = {};

ast.script = function (statements, directives) { return {
    type: stx.Script,
    body: {
        type: stx.FunctionBody,
        directives: directives || [],
        statements: statements || []
    }
}; };

ast.expressionStatement = function (expression) { return {
    type: stx.ExpressionStatement,
    expression: expression || {}
}; };

ast.assignmentExpression = function (binding, expression, operator) { return {
    type: stx.AssignmentExpression,
    operator: operator || '=',
    binding: binding || {},
    expression: expression || {}
}; };

ast.identifierExpression = function (name) { return {
    type: stx.IdentifierExpression,
    identifier: {
        type: stx.Identifier,
        name: name
    }
}; };

ast.objectExpression = function (properties) { return {
    type: stx.ObjectExpression,
    properties: properties || []
}; };

ast.functionExpression = function (statements, name, parameters, directives) {
return {
    type: stx.FunctionExpression,
    name: name || null,
    parameters: parameters || [],
    body: {
        type: stx.FunctionBody,
        directives: directives || [],
        statements: statements || []
    }
}; };

ast.dataProperty = function (name, expression) { return {
    type: stx.DataProperty,
    name: {
        type: stx.PropertyName,
        kind: 'identifier',
        value: name
    },
    expression: expression
}; };

function variate (start, times) {
    var ret = [],
        N = Math.pow(2, start);
    while (times--) {
        ret.push(
            ast.dataProperty(
                'fft' + N,
                ast.functionExpression()
            )
        );
        N *= 2;
    }
    return ret;
}

tree = ast.script([
    ast.expressionStatement(
        ast.assignmentExpression(
            ast.identifierExpression('x'),
            ast.objectExpression(
                variate(4, 4)
            )
        )
    )
]);

// console.log(JSON.stringify(tree, null, 4));
console.log(codegen(tree));
